@@ -14,8 +14,9 @@ from hazelcast.protocol.codec import map_add_entry_listener_codec, map_add_entry
     map_try_put_codec, map_try_remove_codec, map_unlock_codec, map_values_codec, map_values_with_predicate_codec, \
     map_add_interceptor_codec, map_execute_on_all_keys_codec, map_execute_on_key_codec, map_execute_on_keys_codec, \
     map_execute_with_predicate_codec, map_add_near_cache_entry_listener_codec
-from hazelcast.proxy.base import Proxy, EntryEvent, EntryEventType, get_entry_listener_flags
+from hazelcast.proxy.base import Proxy, EntryEvent, EntryEventType, get_entry_listener_flags, MAX_SIZE
 from hazelcast.util import check_not_none, thread_id, to_millis
+from hazelcast import six
 
 
 class Map(Proxy):
@@ -52,13 +53,17 @@ class Map(Proxy):
 
     This class does not allow ``None`` to be used as a key or value.
     """
+    def __init__(self, client, service_name, name):
+        super(Map, self).__init__(client, service_name, name)
+        self.reference_id_generator = self._client.lock_reference_id_generator
+
     def add_entry_listener(self, include_value=False, key=None, predicate=None, added_func=None, removed_func=None, updated_func=None,
                            evicted_func=None, evict_all_func=None, clear_all_func=None, merged_func=None, expired_func=None):
         """
         Adds a continuous entry listener for this map. Listener will get notified for map events filtered with given
         parameters.
 
-        :param include_value: (bool), whether received events include an old value or not (optional).
+        :param include_value: (bool), whether received event should include the value or not (optional).
         :param key: (object), key for filtering the events (optional).
         :param predicate: (Predicate), predicate for filtering the events (optional).
         :param added_func: Function to be called when an entry is added to map (optional).
@@ -296,6 +301,9 @@ class Map(Proxy):
             check_not_none(key, "key can't be None")
             key_list.append(self._to_data(key))
 
+        if len(keys) == 0:
+            return ImmediateFuture([])
+
         return self._encode_invoke(map_execute_on_keys_codec, entry_processor=self._to_data(entry_processor),
                                    keys=key_list)
 
@@ -317,7 +325,8 @@ class Map(Proxy):
         """
         check_not_none(key, "key can't be None")
         key_data = self._to_data(key)
-        return self._encode_invoke_on_key(map_force_unlock_codec, key_data, key=key_data)
+        return self._encode_invoke_on_key(map_force_unlock_codec, key_data, key=key_data,
+                                          reference_id=self.reference_id_generator.get_and_increment())
 
     def get(self, key):
         """
@@ -442,7 +451,7 @@ class Map(Proxy):
         from the server side MapLoader (optional).
         """
         if keys:
-            key_data_list = map(self._to_data, keys)
+            key_data_list = list(map(self._to_data, keys))
             return self._load_all_internal(key_data_list, replace_existing_values)
         else:
             return self._encode_invoke(map_load_all_codec, replace_existing_values=replace_existing_values)
@@ -472,7 +481,9 @@ class Map(Proxy):
         """
         check_not_none(key, "key can't be None")
         key_data = self._to_data(key)
-        return self._encode_invoke_on_key(map_lock_codec, key_data, key=key_data, thread_id=thread_id(), ttl=to_millis(ttl))
+        return self._encode_invoke_on_key(map_lock_codec, key_data, invocation_timeout=MAX_SIZE, key=key_data,
+                                          thread_id=thread_id(), ttl=to_millis(ttl),
+                                          reference_id=self.reference_id_generator.get_and_increment())
 
     def put(self, key, value, ttl=-1):
         """
@@ -513,7 +524,7 @@ class Map(Proxy):
         partition_service = self._client.partition_service
         partition_map = {}
 
-        for key, value in map.iteritems():
+        for key, value in six.iteritems(map):
             check_not_none(key, "key can't be None")
             check_not_none(value, "value can't be None")
             entry = (self._to_data(key), self._to_data(value))
@@ -524,7 +535,7 @@ class Map(Proxy):
                 partition_map[partition_id] = [entry]
 
         futures = []
-        for partition_id, entry_list in partition_map.iteritems():
+        for partition_id, entry_list in six.iteritems(partition_map):
             future = self._encode_invoke_on_partition(map_put_all_codec, partition_id, entries=dict(entry_list))
             futures.append(future)
 
@@ -740,8 +751,9 @@ class Map(Proxy):
 
         key_data = self._to_data(key)
 
-        return self._encode_invoke_on_key(map_try_lock_codec, key_data, key=key_data, thread_id=thread_id(), lease=to_millis(ttl),
-                                          timeout=to_millis(timeout))
+        return self._encode_invoke_on_key(map_try_lock_codec, key_data, invocation_timeout=MAX_SIZE, key=key_data,
+                                          thread_id=thread_id(), lease=to_millis(ttl), timeout=to_millis(timeout),
+                                          reference_id=self.reference_id_generator.get_and_increment())
 
     def try_put(self, key, value, timeout=0):
         """
@@ -786,7 +798,8 @@ class Map(Proxy):
 
         key_data = self._to_data(key)
 
-        return self._encode_invoke_on_key(map_unlock_codec, key_data, key=key_data, thread_id=thread_id())
+        return self._encode_invoke_on_key(map_unlock_codec, key_data, key=key_data, thread_id=thread_id(),
+                                          reference_id=self.reference_id_generator.get_and_increment())
 
     def values(self, predicate=None):
         """
@@ -818,8 +831,8 @@ class Map(Proxy):
     def _get_all_internal(self, partition_to_keys, futures=None):
         if futures is None:
             futures = []
-        for partition_id, key_dict in partition_to_keys.iteritems():
-            future = self._encode_invoke_on_partition(map_get_all_codec, partition_id, keys=key_dict.values())
+        for partition_id, key_dict in six.iteritems(partition_to_keys):
+            future = self._encode_invoke_on_partition(map_get_all_codec, partition_id, keys=list(key_dict.values()))
             futures.append(future)
 
         def merge(f):
@@ -885,16 +898,13 @@ class MapFeatNearCache(Map):
     """
     def __init__(self, client, service_name, name):
         super(MapFeatNearCache, self).__init__(client, service_name, name)
-        near_cache_config = client.config.near_cache_configs.get(name, None)
-        if near_cache_config is None:
-            raise ValueError("NearCache config cannot be None here!")
         self._invalidation_listener_id = None
-        self._near_cache = create_near_cache(client.serialization_service, near_cache_config)
-        if near_cache_config.invalidate_on_change:
+        self._near_cache = client.near_cache_manager.get_or_create_near_cache(name)
+        if self._near_cache.invalidate_on_change:
             self._add_near_cache_invalidation_listener()
 
     def clear(self):
-        self._near_cache.clear()
+        self._near_cache._clear()
         return super(MapFeatNearCache, self).clear()
 
     def evict_all(self):
@@ -933,7 +943,7 @@ class MapFeatNearCache(Map):
         # null key means near cache has to remove all entries in it.
         # see MapAddNearCacheEntryListenerMessageTask.
         if key is None:
-            self._near_cache.clear()
+            self._near_cache._clear()
         else:
             self._invalidate_cache(key)
 
@@ -943,19 +953,11 @@ class MapFeatNearCache(Map):
             self._invalidate_cache(key_data)
 
     def _invalidate_cache(self, key_data):
-        try:
-            del self._near_cache[key_data]
-        except KeyError:
-            # There is nothing to invalidate
-            pass
+        self._near_cache._invalidate(key_data)
 
     def _invalidate_cache_batch(self, key_data_list):
         for key_data in key_data_list:
-            try:
-                del self._near_cache[key_data]
-            except KeyError:
-                # There is nothing to invalidate
-                pass
+            self._near_cache._invalidate(key_data)
 
     # internals
     def _contains_key_internal(self, key_data):
@@ -979,8 +981,8 @@ class MapFeatNearCache(Map):
     def _get_all_internal(self, partition_to_keys, futures=None):
         if futures is None:
             futures = []
-        for key_dic in partition_to_keys.itervalues():
-            for key in key_dic.keys():
+        for key_dic in six.itervalues(partition_to_keys):
+            for key in list(key_dic.keys()):
                 try:
                     key_data = key_dic[key]
                     value = self._near_cache[key_data]
@@ -1048,18 +1050,6 @@ class MapFeatNearCache(Map):
         return super(MapFeatNearCache, self)._delete_internal(key_data)
 
 
-def create_near_cache(serialization_service, near_cache_config):
-    return NearCache(serialization_service,
-                     near_cache_config.in_memory_format,
-                     near_cache_config.time_to_live_seconds,
-                     near_cache_config.max_idle_seconds,
-                     near_cache_config.invalidate_on_change,
-                     near_cache_config.eviction_policy,
-                     near_cache_config.eviction_max_size,
-                     near_cache_config.eviction_sampling_count,
-                     near_cache_config.eviction_sampling_pool_size)
-
-
 def create_map_proxy(client, service_name, name, **kwargs):
     near_cache_config = client.config.near_cache_configs.get(name, None)
     if near_cache_config is None:
